package fr.mrdev023.vulkan_java.vk;

import static fr.mrdev023.vulkan_java.vk.VulkanError.vkCheck;
import static org.lwjgl.vulkan.VK10.VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
import static org.lwjgl.vulkan.VK10.vkCreateDevice;
import static org.lwjgl.vulkan.VK10.vkDestroyDevice;
import static org.lwjgl.vulkan.VK10.vkDeviceWaitIdle;

import java.util.Optional;

import org.lwjgl.PointerBuffer;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.VkDevice;
import org.lwjgl.vulkan.VkDeviceCreateInfo;
import org.lwjgl.vulkan.VkPhysicalDeviceFeatures;
import org.tinylog.Logger;

import fr.mrdev023.vulkan_java.vk.utils.SuitablePhysicalDeviceFinder;

public class Device {

    private final PhysicalDevice physicalDevice;
    private final VkDevice vkDevice;
    private final Optional<Queue> presentQueue;
    private final Optional<Queue> graphicsQueue;
    private final Optional<Queue> computeQueue;
    private final Optional<Queue> transferQueue;

    public Device(SuitablePhysicalDeviceFinder.MatchResult physicalDeviceMatch) throws VulkanError {
        this.physicalDevice = physicalDeviceMatch.physicalDevice;

        try (MemoryStack stack = MemoryStack.stackPush()) {
            var extensions = physicalDeviceMatch.validatorResult.deviceExtensionsResult.matchingExtensions;
            PointerBuffer requiredExtensions = stack.mallocPointer(extensions.size());
            for (String extension : extensions) {
                requiredExtensions.put(stack.ASCII(extension));
            }
            requiredExtensions.flip();

            // Set up required features
            VkPhysicalDeviceFeatures features = VkPhysicalDeviceFeatures.calloc(stack);

            // Enable all the queue families
            var queueFamilyIndices = physicalDeviceMatch.validatorResult.queueFamilyIndices.orElseThrow();
            var queuesCreator = new QueuesCreator(queueFamilyIndices);
            var queueCreationInfoBuf = queuesCreator.createQueueCreateInfos(stack, physicalDevice);

            VkDeviceCreateInfo deviceCreateInfo = VkDeviceCreateInfo.calloc(stack)
                    .sType(VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO)
                    .ppEnabledExtensionNames(requiredExtensions)
                    .pEnabledFeatures(features)
                    .pQueueCreateInfos(queueCreationInfoBuf);

            PointerBuffer pp = stack.mallocPointer(1);
            vkCheck(vkCreateDevice(physicalDevice.getVkPhysicalDevice(), deviceCreateInfo, null, pp),
                    "Failed to create device");
            vkDevice = new VkDevice(pp.get(0), physicalDevice.getVkPhysicalDevice(), deviceCreateInfo);

            graphicsQueue = queueFamilyIndices.getGraphicsQueueFamilyIndex()
                    .flatMap(queueFamilyIndex -> Optional.of(queuesCreator.createQueue(this, queueFamilyIndex)));
            presentQueue = queueFamilyIndices.getPresentQueueFamilyIndex()
                    .flatMap(queueFamilyIndex -> Optional.of(queuesCreator.createQueue(this, queueFamilyIndex)));
            computeQueue = queueFamilyIndices.getComputeQueueFamilyIndex()
                    .flatMap(queueFamilyIndex -> Optional.of(queuesCreator.createQueue(this, queueFamilyIndex)));
            transferQueue = queueFamilyIndices.getTransferQueueFamilyIndex()
                    .flatMap(queueFamilyIndex -> Optional.of(queuesCreator.createQueue(this, queueFamilyIndex)));
        }

        Logger.debug("Vulkan device created");
    }

    public void destroy() {
        Logger.debug("Destroying Vulkan device");
        vkDestroyDevice(vkDevice, null);
    }

    public PhysicalDevice getPhysicalDevice() {
        return physicalDevice;
    }

    public VkDevice getVkDevice() {
        return vkDevice;
    }

    public void waitIdle() {
        vkDeviceWaitIdle(vkDevice);
    }

    public Optional<Queue> getPresentQueue() {
        return presentQueue;
    }

    public Optional<Queue> getGraphicsQueue() {
        return graphicsQueue;
    }

    public Optional<Queue> getComputeQueue() {
        return computeQueue;
    }

    public Optional<Queue> getTransferQueue() {
        return transferQueue;
    }
}