package fr.mrdev023.vulkan_java.vk.utils;

import java.nio.IntBuffer;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collectors;

import org.lwjgl.PointerBuffer;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.VK10;
import org.lwjgl.vulkan.VkLayerProperties;
import org.tinylog.Logger;

public class InstanceValidationLayers {
    private Set<String> validationLayers;
    
    private InstanceValidationLayers(Set<String> validationLayers) {
        this.validationLayers = validationLayers;
    }

    public boolean hasValidationLayers() {
        return !validationLayers.isEmpty();
    }

    public Set<String> getValidationLayers() {
        return validationLayers;
    }

    public PointerBuffer writeToStack(MemoryStack stack) {
        if (validationLayers.isEmpty()) {
            return null;
        }

        PointerBuffer requiredLayers = stack.mallocPointer(validationLayers.size());
        for (String layer : validationLayers) {
            requiredLayers.put(stack.UTF8(layer));
        }
        requiredLayers.flip();

        return requiredLayers;
    }

    /**
     * This class is used to select the validation layers and create an InstanceValidationLayers object.
     * 
     * <h5>Specification</h5>
     * <ul>
     *  <li>Include the validation layers if the user wants them</li>
     *  <li>Select the validation layers from the supported validation layers</li>
     *  <li>Return an InstanceValidationLayers object with the selected validation layers</li>
     * </ul>
     * 
     * <h5>See Also</h5>
     * <ul>
     *  <li>{@link #includeValidationLayers(boolean)}</li>
     *  <li>{@link #selectValidationLayers()}</li>
     * </ul>
     */
    public static class Selector {
        private boolean includeValidationLayers = false;

        public Selector includeValidationLayers(boolean includeValidationLayers) {
            this.includeValidationLayers = includeValidationLayers;
            return this;
        }

        /**
         * Selects the validation layers.
         * 
         * @return the set of validation layers
         */
        public InstanceValidationLayers selectValidationLayers() {
            var supportedLayers = getSupportedValidationLayers();
            log("Supported validation layers", supportedLayers);

            Set<String> layersToUse = includeValidationLayers
                ? selectValidationLayersFromSupportedLayers(supportedLayers)
                : Collections.emptySet();
            log("Selected validation layers", layersToUse);

            if (layersToUse.isEmpty() && includeValidationLayers) {
                Logger.warn("Request validation but no supported validation layers found. Falling back to no validation");
            }

            return new InstanceValidationLayers(layersToUse);
        }

        private Set<String> getSupportedValidationLayers() {
            Set<String> supportedLayers = new HashSet<>();

            try (MemoryStack stack = MemoryStack.stackPush()) {
                // Get the number of supported validation layers by the current vulkan instance
                IntBuffer numLayersArr = stack.callocInt(1);
                VK10.vkEnumerateInstanceLayerProperties(numLayersArr, null);
                int numLayers = numLayersArr.get(0);

                // Get the supported validation layers by the current vulkan instance
                VkLayerProperties.Buffer propsBuf = VkLayerProperties.calloc(numLayers, stack);
                VK10.vkEnumerateInstanceLayerProperties(numLayersArr, propsBuf);

                // Convert the supported validation layers to a list of strings
                for (int i = 0; i < numLayers; i++) {
                    VkLayerProperties props = propsBuf.get(i);
                    String layerName = props.layerNameString();
                    supportedLayers.add(layerName);
                }
            }

            return supportedLayers;
        }

        private Set<String> selectValidationLayersFromSupportedLayers(Set<String> supportedLayers) {
            Set<String> layersToUse = new HashSet<>();

            // Main validation layer
            if (supportedLayers.contains("VK_LAYER_KHRONOS_validation")) {
                layersToUse.add("VK_LAYER_KHRONOS_validation");
                return layersToUse;
            }

            // Fallback 1
            if (supportedLayers.contains("VK_LAYER_LUNARG_standard_validation")) {
                layersToUse.add("VK_LAYER_LUNARG_standard_validation");
                return layersToUse;
            }

            // Fallback 2 (set)
            Set<String> requestedLayers = new HashSet<>();
            requestedLayers.add("VK_LAYER_GOOGLE_threading");
            requestedLayers.add("VK_LAYER_LUNARG_parameter_validation");
            requestedLayers.add("VK_LAYER_LUNARG_object_tracker");
            requestedLayers.add("VK_LAYER_LUNARG_core_validation");
            requestedLayers.add("VK_LAYER_GOOGLE_unique_objects");

            return requestedLayers.stream().filter(supportedLayers::contains).collect(Collectors.toSet());
        }

        private void log(String title, Set<String> layers) {
            Logger.debug("{} ({})", title, layers.size());
            for (String layer : layers) {
                Logger.debug(" - {}", layer);
            }
        }
    }
}