package fr.mrdev023.vulkan_java.vk.validators;

import java.util.HashSet;
import java.util.Set;

import fr.mrdev023.vulkan_java.vk.PhysicalDevice;

public class DeviceExtensionsValidator {
    private final Set<String> requiredExtensions;

    public DeviceExtensionsValidator(Set<String> requiredExtensions) {
        this.requiredExtensions = requiredExtensions;
    }

    /**
     * Validate the device extensions of the physical device.
     * 
     * <i>Note: This method return a Result to adapt the extensions to the physical
     * device properties after the validation.</i>
     * 
     * @param physicalDevice The physical device to validate the extensions for.
     * @return A result object containing the validation results and whether the
     *         physical device is suitable.
     */
    public Result validate(PhysicalDevice physicalDevice) {
        Set<String> matchingExtensions = new HashSet<>();
        Set<String> missingExtensions = new HashSet<>();

        var vkDeviceExtensions = physicalDevice.getVkDeviceExtensions();

        for (var extension : requiredExtensions) {
            var hasExtension = vkDeviceExtensions.stream().anyMatch(e -> e.extensionNameString().equals(extension));
            if (hasExtension) {
                matchingExtensions.add(extension);
            } else {
                missingExtensions.add(extension);
            }
        }

        return new Result(matchingExtensions, missingExtensions, missingExtensions.isEmpty());
    }

    public static class Result {
        public final Set<String> matchingExtensions;
        public final Set<String> missingExtensions;
        public final boolean isSuitable;

        public Result(Set<String> matchingExtensions, Set<String> missingExtensions, boolean isSuitable) {
            this.matchingExtensions = matchingExtensions;
            this.missingExtensions = missingExtensions;
            this.isSuitable = isSuitable;
        }
    }
}
