package fr.mrdev023.vulkan_java.vk.validators;

import java.nio.IntBuffer;
import java.util.Optional;

import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.KHRSurface;
import org.lwjgl.vulkan.VK10;
import org.lwjgl.vulkan.VkQueueFamilyProperties;

import fr.mrdev023.vulkan_java.vk.PhysicalDevice;
import fr.mrdev023.vulkan_java.vk.QueueFamilyIndices;
import fr.mrdev023.vulkan_java.vk.Surface;

public class QueueFamilyRequirementsValidator {
    private int requiredQueueFlags;
    private Optional<Surface> surface;

    public QueueFamilyRequirementsValidator(int requiredQueueFlags, Optional<Surface> surface) {
        this.requiredQueueFlags = requiredQueueFlags;
        this.surface = surface;
    }

    /**
     * Validate the queue family requirements of the physical device.
     * 
     * @param physicalDevice The physical device to validate the requirements for.
     * @return The queue family indices if the requirements are met, otherwise an
     *         empty optional.
     */
    public Optional<QueueFamilyIndices> validate(PhysicalDevice physicalDevice) {
        var vkQueueFamilyProps = physicalDevice.getVkQueueFamilyProps();

        var queueFamilyIndicesBuilder = new QueueFamilyIndices.Builder();

        for (int i = 0; i < vkQueueFamilyProps.capacity(); i++) {
            var vkQueueFamilyProp = vkQueueFamilyProps.get(i);

            if (hasRequiredFlag(VK10.VK_QUEUE_GRAPHICS_BIT) && queueFlagIsSupported(vkQueueFamilyProp)) {
                queueFamilyIndicesBuilder.withGraphicsQueueFamilyIndex(i);
                requiredQueueFlags &= ~VK10.VK_QUEUE_GRAPHICS_BIT; // Consume the flag
            }

            if (hasRequiredFlag(VK10.VK_QUEUE_COMPUTE_BIT) && queueFlagIsSupported(vkQueueFamilyProp)) {
                queueFamilyIndicesBuilder.withComputeQueueFamilyIndex(i);
                requiredQueueFlags &= ~VK10.VK_QUEUE_COMPUTE_BIT; // Consume the flag
            }

            if (hasRequiredFlag(VK10.VK_QUEUE_TRANSFER_BIT) && queueFlagIsSupported(vkQueueFamilyProp)) {
                queueFamilyIndicesBuilder.withTransferQueueFamilyIndex(i);
                requiredQueueFlags &= ~VK10.VK_QUEUE_TRANSFER_BIT; // Consume the flag
            }

            if (surface.isPresent() && isSurfaceSupported(surface.get(), physicalDevice, i)) {
                queueFamilyIndicesBuilder.withPresentQueueFamilyIndex(i);
                surface = Optional.empty(); // Consume the surface
            }

            if (requiredQueueFlags == 0 && surface.isEmpty()) {
                break;
            }
        }

        if (requiredQueueFlags == 0 && surface.isEmpty()) {
            return Optional.of(queueFamilyIndicesBuilder.build());
        }

        return Optional.empty();
    }

    private boolean queueFlagIsSupported(VkQueueFamilyProperties vkQueueFamilyProp) {
        return (vkQueueFamilyProp.queueFlags() & requiredQueueFlags) != 0;
    }

    private boolean isSurfaceSupported(Surface surface, PhysicalDevice physicalDevice, int queueFamilyIndex) {
        try (MemoryStack stack = MemoryStack.stackPush()) {
            IntBuffer presentSupport = stack.ints(VK10.VK_FALSE);
            KHRSurface.vkGetPhysicalDeviceSurfaceSupportKHR(physicalDevice.getVkPhysicalDevice(), queueFamilyIndex,
                    surface.getVkSurface(), presentSupport);
            return presentSupport.get(0) == VK10.VK_TRUE;
        }
    }

    private boolean hasRequiredFlag(int flag) {
        return (requiredQueueFlags & flag) != 0;
    }
}
